########################################################################
# Integração Numérica
########################################################################


## aproximação por (i) função escada e (ii) quadraturas

##
## 1. Integração Unidimensional
##

## Exemplo:
# função: \int_0^1 exp(x) dx

## 1.1 solução analítica:
(I <- exp(1) - 1)

## 1.2 Função disponível no R:
##     solução por integrate() (uma dimensão e limitada)
I.int <- integrate(function(x) exp(x), lower=0, upper=1)
I.int
## resultado aproximado, exato, erro e erro relativo
c(I.int$val, I, I-I.int$val, (I-I.int$val)/I)

## 1.3 approximação por ponto médio
N <- 10
I.pm <- sum(exp((1:N-.5)/N))/N
I.pm
c(I.pm, I, (I-I.pm)/I)

## pode-se aumentar a acurácia...
N <- 100
I.pm <- sum(exp((1:N-.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

N <- 1000
I.pm <- sum(exp((1:N -.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

## 1.4 quadratura gaussiana
require(statmod)
N <- 10
## pontos x_i de integração e pesos w_i
qg <- gauss.quad(N) 
qg
## a função é reescrita para aplicar a quadratura e no intervalo [-1,1]
## usando a relação \int^a_b f(x) dx = \int_{-1}^1 f(/frac{a+b+(b-a)t}{2}) dt
I.qg <- sum(qg$weights * exp((qg$nodes+1)/2))/2
c(I.qg,I,(I-I.qg)/I)

## reduzindo o número de pontos...
N <- 5
gq <- gauss.quad(N) 
I.qg <- sum(gq$weights*exp((gq$nodes+1)/2))/2
c(I.qg, I, (I-I.qg)/I)

N <- 3
qg <- gauss.quad(N) 
I.qg <- sum(qg$weights*exp((qg$nodes+1)/2))/2
c(I.qg, I, (I - I.qg)/I)

## examinar os argumentos da função para outras opções e métodos! 

## 1.5 Outros tipos/opcoes de quadratura

## outras dadas opções são ainda dadas em gauss.quad.prob()
args(gauss.quad.prob)

## uniforme
qgu <- gauss.quad.prob(10, dist="uniform", l=0, u=1)
qgu
I.qgu <- sum(qgu$w * exp(qgu$n))
I.qgu
c(I.qgu, I, (I - I.qgu)/I)

## normal (não é muito natural para este exemplo, pode-se tentar
## diferentes opções para mu e sigma mas aproximação não funciona muito bem...)
## Tem-se aqui o problema do domínio de integração estar restrito de em [0,1] 
## Esta opção é mais adequada para variáveis como domínio de integração entre (-Inf, Inf) 
## e requer que mu e sigma sejam informados
qgn <- gauss.quad.prob(10, dist="normal", mu=0.5, sigma=0.2)
qgn

## idem anterior para Gamma com parâmetro de forma alpha e escala beta
## só que neste caso com intervalo (0, +Inf)
qgg <- gauss.quad.prob(10, dist="gamma", alpha=1, beta=0.5)
qgg
## e note que para alpha (shape) = 1 tem-se a exponencial
ff <- function(x) dgamma(x, sh=1, sc=1/5)
curve(ff, from=0, to=3)
#
dgamma(0:5, sh=1, rate=1/10)
dexp(0:5, rate=1/10)

## Outra alternativa para o intervalo (0, 1) é dado pela Beta
##
bb <- function(x, a, b) x^(a-1)*(1-x)^(b-1)/beta(a,b)
bb((0:10)/10, 2, 3)
dbeta((0:10)/10, 2, 3)
ff <- function(x) dbeta(x, 2, 2)
curve(ff, 0, 1)
ff <- function(x) dbeta(x, 0.5, 0.5)
curve(ff, 0, 1, add=T, lty=2)
ff <- function(x) dbeta(x, 1.5, 0.5)
curve(ff, 0, 1, add=T, lwd=2)
ff <- function(x) dbeta(x, 1, 1)
curve(ff, 0, 1, add=T, lty=2, lwd=2)

## mas note que a qualidade a aproximação depende da escolha dos parâmetros
qgb <- gauss.quad.prob(10, dist="beta", a=2, b=2)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=0.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1, b=1)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

##
## 2. Integração Multidimensional...
##

## Exemplo:

## f(x) = exp(\sum x_i), em [-1; 1]^d
fd <- function(x) {exp(rowSums(x))}

## 2.1 Solução analítica
I <- (exp(1)-exp(-1))^d

## 2.2 Funções disponíveis no R
## verificar "adapt" e outras opções (ver CRAN Task Views - NumericalMathematics)

## Aproximações numéricas:
## baseadas na aplicação recursiva de regras de 1-D
## \int \int f(x,y) dx dy \approx \sum_j \sum_i w_j w_i f(x_j, y_i)

## função para encontrar conjunto de nós e pesos d-dimensional,
malha <- function(nos, dim, pesos = 1/length(x)+x*0) {
  ## nos: nós (em 1-D) onde a função deve ser avaliada
  ## dim: dimensão do espaço de integração
  ## pesos: pesos correspondentes aos nós 
  n <- length(nos)
  X <- W <- matrix(0, n^dim, dim)
  for (i in 1:dim) {
    X[,i] <- nos
    W[,i] <- pesos
    nos <- rep(nos, rep(n, length(nos)))
    pesos <- rep(pesos, rep(n, length(pesos)))
  }
  ## pesos são dados pelo produto das colunas de W
  pesos <- exp(rowSums(log(W))) 
  ## e cada linha de X fornece as coordenadas dos nós
  list(nodes=X, weights=pesos) 
}

N <- 10; d <- 3
## número de avaliações da função:
N^d

## 2.3 Aproximação por ponto médio:
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$weights * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)

## agora em 6-D 
N <- 10; d <- 6
## número de avaliações da função
N^d

I <- (exp(1)-exp(-1))^d
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$wei * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)

## 2.4 Aproximação por quadratura

## quadratura gaussiana, aumenta (mas nem tanto!, para uma mesma precisao) o número de avaliações da função
N <- 5; d <- 2
N^d

I <- (exp(1)-exp(-1))^d
gq <- gauss.quad(N)
malha.qg <- malha(gq$nodes, d, gq$weights)
I.qg <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)

## comparar o efeito do aumento da dimensão
N <- 10; d <- 6
N^d

I <- (exp(1)-exp(-1))^d
qg <- gauss.quad(N)
malha.gq <- malha(gq$nodes, d, gq$weights)
I.gq <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)

## experimentar com outros números de nós etc etc

##
## Algumas outras abordagens para aproximação:                                                     
##

## Aproximando o integrando: abordagem por Aproximação de Laplace
## Integração de Monte Carlo
## Monte Carlo estratificado
## Amostragem por importância (Importance sampling)
## Amostragem por importância via Laplace (Laplace importance sampling)



